﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Analysis.SolutionCaching.RealVectorEncoding {
  [Item("RealVectorCrossoverPointcut", "Wraps a crossover and applies after-crossover actions.")]
  [StorableClass]
  public class RealVectorCrossoverPointcut : SingleSuccessorOperator, IRealVectorCrossover, IStochasticOperator, IRealVectorPointcut {
    private const string AdviceParameterName = "Advice";
    private const string CrossoversParameterName = "Crossovers";

    public ConstrainedValueParameter<IRealVectorCrossover> CrossoversParameter {
      get { return (ConstrainedValueParameter<IRealVectorCrossover>)Parameters[CrossoversParameterName]; }
    }
    public ConstrainedValueParameter<IRealVectorAdvice> AdviceParameter {
      get { return (ConstrainedValueParameter<IRealVectorAdvice>)Parameters[AdviceParameterName]; }
    }
    public ILookupParameter<ItemArray<RealVector>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<RealVector>)Parameters["Parents"]; }
    }
    public ILookupParameter<RealVector> ChildParameter {
      get { return (ILookupParameter<RealVector>)Parameters["Child"]; }
    }
    public IValueLookupParameter<DoubleMatrix> BoundsParameter {
      get { return (IValueLookupParameter<DoubleMatrix>)Parameters["Bounds"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    [StorableConstructor]
    protected RealVectorCrossoverPointcut(bool deserializing) : base(deserializing) { }
    protected RealVectorCrossoverPointcut(RealVectorCrossoverPointcut original, Cloner cloner) : base(original, cloner) { }

    public RealVectorCrossoverPointcut()
      : base() {
      var cxs = new ConstrainedValueParameter<IRealVectorCrossover>(CrossoversParameterName, "The crossovers.");
      //Would be nicer if it comes from operator collection, but MultiRealVector Crossover does the same thing.
      //Also: MultiRealVectorCrossover cannot be used here as it would generate a cyclic recursion 
      //MultiRealVectorCrossover discovers RealVectorCrossoverPointcut which discovers Multi....
      foreach (Type t in ApplicationManager.Manager.GetTypes(typeof(IRealVectorCrossover))) {
        if ((!typeof(RealVectorCrossoverPointcut).IsAssignableFrom(t)) &&
          (!typeof(MultiOperator<IRealVectorCrossover>).IsAssignableFrom(t))) {
          cxs.ValidValues.Add((IRealVectorCrossover)Activator.CreateInstance(t));
        }
      }
      Parameters.Add(cxs);

      var advices = new ConstrainedValueParameter<IRealVectorAdvice>(AdviceParameterName,
        "Advice to introduce after crossover");
      foreach (var a in ApplicationManager.Manager.GetInstances<IRealVectorAdvice>()) {
        advices.ValidValues.Add(a);
      }
      Parameters.Add(advices);

      Parameters.Add(new ScopeTreeLookupParameter<RealVector>("Parents", "The parent vectors which should be crossed."));
      ParentsParameter.ActualName = "RealVector";
      Parameters.Add(new LookupParameter<RealVector>("Child", "The child vector resulting from the crossover."));
      ChildParameter.ActualName = "RealVector";
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("Bounds", "The lower and upper bounds of the real vector."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));

      ParameterizeCrossovers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RealVectorCrossoverPointcut(this, cloner);
    }

    public override IOperation Apply() {
      IOperation successor = base.Apply();
      IOperation crossoverOperation = ExecutionContext.CreateChildOperation(CrossoversParameter.Value);
      IOperation adviceOperation = ExecutionContext.CreateChildOperation(AdviceParameter.Value);

      var opCol = new OperationCollection();
      opCol.Add(crossoverOperation);
      opCol.Add(adviceOperation);
      if (successor != null)
        opCol.Add(successor);

      return opCol;
    }

    private void ParameterizeCrossovers() {
      foreach (IRealVectorCrossover crossover in CrossoversParameter.ValidValues.OfType<IRealVectorCrossover>()) {
        crossover.ChildParameter.ActualName = ChildParameter.Name;
        crossover.ParentsParameter.ActualName = ParentsParameter.Name;
        crossover.BoundsParameter.ActualName = BoundsParameter.Name;
      }
      foreach (IStochasticOperator crossover in CrossoversParameter.ValidValues.OfType<IStochasticOperator>()) {
        crossover.RandomParameter.ActualName = RandomParameter.Name;
      }
    }
  }
}
